"""
Defines two classes that help with balanced sampling (with replacement)
"""
import numpy as np
import ubelt as ub
import kwarray
import networkx as nx


try:
    from line_profiler import profile
except Exception:
    profile = ub.identity


class BalancedSampleTree(ub.NiceRepr):
    """
    Manages a sampling from a tree of indexes. Helps with balancing
    samples over multiple criteria.

    TODO:
        Move to its own file - possibly a new module. This is a very general
        construct, and would benefit from binary-language optimizations.

    Example:
        >>> from geowatch.tasks.fusion.datamodules.balanced_sampling import BalancedSampleTree
        >>> # Given a grid of sample locations and attribute information
        >>> # (e.g., region, category).
        >>> sample_grid = [
        >>>     { 'region': 'region1', 'category': 'background', 'color': "blue" },
        >>>     { 'region': 'region1', 'category': 'background', 'color': "purple" },
        >>>     { 'region': 'region1', 'category': 'background', 'color': "blue" },
        >>>     { 'region': 'region1', 'category': 'background', 'color': "red" },
        >>>     { 'region': 'region1', 'category': 'background', 'color': "green" },
        >>>     { 'region': 'region1', 'category': 'background', 'color': "purple" },
        >>>     { 'region': 'region1', 'category': 'background', 'color': "blue" },
        >>>     { 'region': 'region1', 'category': 'rare',       'color': "red" },
        >>>     { 'region': 'region1', 'category': 'rare',       'color': "green" },
        >>>     { 'region': 'region1', 'category': 'background', 'color': "red" },
        >>>     { 'region': 'region1', 'category': 'background', 'color': "green" },
        >>>     { 'region': 'region2', 'category': 'background', 'color': "blue" },
        >>>     { 'region': 'region2', 'category': 'background', 'color': "purple" },
        >>>     { 'region': 'region2', 'category': 'background', 'color': "red" },
        >>>     { 'region': 'region2', 'category': 'background', 'color': "green" },
        >>>     { 'region': 'region2', 'category': 'rare',       'color': "purple" },
        >>>     { 'region': 'region2', 'category': 'rare',       'color': "blue" },
        >>> ]
        >>> #
        >>> # First we can just create a flat uniform sampling grid
        >>> # and inspect the imbalance that causes.
        >>> self = BalancedSampleTree(sample_grid, rng=0)
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist0 = ub.dict_hist([(g['region'], g['category']) for g in sampled])
        >>> print('hist0 = {}'.format(ub.urepr(hist0, nl=1)))
        >>> #
        >>> # We can subdivide the indexes based on region to improve balance.
        >>> self.subdivide('region')
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist1 = ub.dict_hist([(g['region'], g['category']) for g in sampled])
        >>> print('hist1 = {}'.format(ub.urepr(hist1, nl=1)))
        >>> #
        >>> # We can further subdivide by category.
        >>> self.subdivide('category')
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist2 = ub.dict_hist([(g['region'], g['category']) for g in sampled])
        >>> print('hist2 = {}'.format(ub.urepr(hist2, nl=1)))
        >>> #
        >>> # We can further subdivide by color, with custom weights.
        >>> weights = { 'red': .25, 'blue': .25, 'green': .4, 'purple': .1 }
        >>> self.subdivide('color', weights=weights)
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist3 = ub.dict_hist([
        >>>     (g['region'], g['category'], g['color']) for g in sampled
        >>> ])
        >>> print('hist3 = {}'.format(ub.urepr(hist3, nl=1)))
        >>> hist3_color = ub.dict_hist([(g['color']) for g in sampled])
        >>> print('color weights = {}'.format(ub.urepr(weights, nl=1)))
        >>> print('hist3 (color) = {}'.format(ub.urepr(hist3_color, nl=1)))
    """
    @profile
    def __init__(self, sample_grid, rng=None):
        """
        Args:
            sample_grid (List[Dict]):
                List of items with properties to be sampled

            rng (int | None | RandomState):
                random number generator or seed
        """
        self.rng = kwarray.ensure_rng(rng)

        # validate input
        if not isinstance(sample_grid, list):
            raise TypeError(ub.paragraph(
                """
                BalancedSampleTree only accepts List[Dict], but outer type
                was {type(sample_grid)}.
                """))

        if not sample_grid:
            raise ValueError('Input sample_grid is empty')

        if not isinstance(sample_grid[0], dict):
            raise TypeError(ub.paragraph(
                """
                BalancedSampleTree only accepts List[Dict], but inner type
                was {type(sample_grid[0])}.
                """))

        self.graph = self._create_graph(sample_grid)
        self._leaf_nodes = [n for n in self.graph.nodes if self.graph.out_degree[n] == 0]

    def reseed(self, rng):
        """
        Reseed (or unseed) the random number generator

        Args:
            rng (int | None | RandomState):
                random number generator or seed
        """
        self.rng = kwarray.ensure_rng(rng)

    @profile
    def _create_graph(self, sample_grid):
        graph = nx.DiGraph()

        # make a special root node
        root_node = '__root__'
        graph.add_node(root_node, weights=None)

        for index, item in enumerate(sample_grid):
            # Using urepr in the critial loop is too slow for large sample
            # grids
            # maybe we add an option to enable this for debugging / demo?
            # label = f'{index:02d} ' + ub.urepr(item, nl=0, compact=1, nobr=1)
            label = f'{index:02d}'
            graph.add_node(index, label=label, **item)
            graph.add_edge(root_node, index)
        return graph

    @profile
    def _get_parent(self, n):
        """ Get the parent of a node (assume a tree). None if it doesnt exist """
        preds = self.graph.pred[n]
        # This function is called a lot, disable sanity checks
        # if len(preds):
        #     assert len(preds) == 1
        #     return next(iter(preds))
        # else:
        #     return None
        return next(iter(preds))

    @profile
    def _reweight(self, node, idx_child):
        if self.graph.nodes[node]['weights'] is not None:
            _weights = self.graph.nodes[node]['weights']

            # remove weight for this child
            _weights = np.delete(_weights, idx_child)

            # reweight
            if _weights.sum() != 0:
                _weights = _weights / _weights.sum()
            else:
                _weights = np.zeros(1)
            self.graph.nodes[node]['weights'] = _weights

    @profile
    def _prune_and_reweight(self, nodes):
        for parent, orphans in nodes:
            grandpa = self._get_parent(parent)
            if grandpa is None:
                # already removed this branch
                self.graph.remove_nodes_from([parent] + orphans)
                continue

            # get parent index from grandpa, remove nodes
            idx_parent = list(self.graph.successors(grandpa)).index(parent)
            self.graph.remove_nodes_from([parent] + orphans)

            # update weights of grandpa, walking up the tree
            queue = [(grandpa, idx_parent)]
            while queue:
                curr_grandpa, curr_idx_parent = queue.pop()

                num_children = len(list(self.graph.successors(curr_grandpa)))
                if num_children >= 1:
                    self._reweight(curr_grandpa, curr_idx_parent)
                else:
                    # removed only child, remove the grandparent
                    _parent = curr_grandpa
                    _grandpa = self._get_parent(curr_grandpa)
                    if _grandpa is not None:
                        _idx_parent = list(self.graph.successors(_grandpa)).index(_parent)
                        queue.append((_grandpa, _idx_parent))
                        self.graph.remove_node(curr_grandpa)

        # update leaf nodes
        self._leaf_nodes = [n for n in self._leaf_nodes if self.graph.has_node(n)]
        if len(self._leaf_nodes) == 0:
            raise ValueError("Leaf nodes became empty.")

    @profile
    def subdivide(self, key, weights=None, default_weight=0):
        """
        Adds a new layer to the tree that balances across the given attribute.

        Args:
            key (str):
                A key into the item dictionary of a sample that maps to the
                property to balance over.

            weights (None | Dict[Any, Number]):
                an optional mapping from values that ``key`` could point to
                to a numeric weight.

            default_weight (None | Number):
                if an attribute is unspecified in the weight table, this is
                the default weight it should be given. Default is 0.
        """
        remove_nodes = []
        remove_edges = []
        add_edges = []
        add_nodes = []

        # Group all leaf nodes by their direct parents

        # It is possible that we could optimize this with a column-based data
        # structure, but this current structure if far more general and easier
        # to read.
        parent_to_leafs = ub.group_items(self._leaf_nodes, key=lambda n: self._get_parent(n))

        for parent, children in parent_to_leafs.items():
            # Group children by the new attribute
            val_to_subgroup = ub.group_items(children, lambda n: self.graph.nodes[n][key])
            # try:
            #     val_to_subgroup = ub.odict(sorted(val_to_subgroup.items()))
            # except TypeError:
            #     val_to_subgroup = ub.odict(sorted(val_to_subgroup.items(), key=str))

            # Add weights to the prior parent
            if weights is not None:
                weights_group = np.asarray(list(ub.take(weights, val_to_subgroup.keys(), default=default_weight)))
                denom = weights_group.sum()
                if denom != 0:
                    weights_group = weights_group / denom
                    self.graph.nodes[parent]['weights'] = weights_group
                else:
                    # All options have zero weight, schedule group for pruning
                    remove_nodes.append((parent, children))
                    continue
            else:
                self.graph.nodes[parent]["weights"] = None

            # Create a node for each child
            for value, subgroup in val_to_subgroup.items():
                # Use a dotted name to make unambiguous tree splits
                new_parent = f'{parent}.{key}={value}'
                # Mark edges to add / remove to implement the split
                remove_edges.extend([(parent, n) for n in subgroup])
                add_edges.extend([(parent, new_parent) for n in subgroup])
                add_edges.extend([(new_parent, n) for n in subgroup])
                add_nodes.append(new_parent)

        # Modify the graph
        self.graph.remove_edges_from(remove_edges)
        self.graph.add_nodes_from(add_nodes, weights=None)
        self.graph.add_edges_from(add_edges)
        self._prune_and_reweight(remove_nodes)

    @profile
    def _sample_many(self, num):
        for _ in range(num):
            idx = self.sample()
            yield idx

    @profile
    def sample(self):
        """
        Returns:
            int
        """
        current = '__root__'
        while self.graph.out_degree(current) > 0:
            children = list(self.graph.successors(current))
            num = len(children)

            weights = self.graph.nodes[current]['weights']
            if weights is None:
                idx = self.rng.randint(0, num)
            else:
                idx = self.rng.choice(num, 1, p=weights)[0]

            current = children[idx]
        return current

    @profile
    def __len__(self):
        return len(list(self._leaf_nodes))

    @profile
    def __nice__(self):
        n_nodes = self.graph.number_of_nodes()
        n_edges = self.graph.number_of_edges()
        n_leafs = self.__len__()
        n_depth = len(nx.algorithms.dag.dag_longest_path(self.graph))
        return f'nodes={n_nodes}, edges={n_edges}, leafs={n_leafs}, depth={n_depth}'


class BalancedSampleForest(ub.NiceRepr):
    """
    Manages a sampling from a forest of BalancedSampleTree's. Helps with balancing
    samples in the multi-label case.

    CommandLine:
        LINE_PROFILE=1 xdoctest -m geowatch.tasks.fusion.datamodules.balanced_sampling BalancedSampleForest:1 --benchmark

    Example:
        >>> from geowatch.tasks.fusion.datamodules.balanced_sampling import BalancedSampleForest
        >>> sample_grid = [
        >>>     { 'region': 'region1', 'color': {'blue': 10, 'red': 3}},
        >>>     { 'region': 'region1', 'color': {'green': 3, 'purple': 2}},
        >>>     { 'region': 'region1', 'color': {'blue': 1}},
        >>>     { 'region': 'region1', 'color': {'green': 3, 'red': 5}},
        >>>     { 'region': 'region1', 'color': {'purple': 1, 'blue': 1}},
        >>>     { 'region': 'region2', 'color': {'blue': 5, 'red': 5}},
        >>>     { 'region': 'region2', 'color': {'green': 5, 'purple': 5}},
        >>> ]
        >>> #
        >>> self = BalancedSampleForest(sample_grid, rng=0)
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist0 = ub.dict_hist([g['region'] for g in sampled])
        >>> print('hist0 = {}'.format(ub.urepr(hist0, nl=1)))
        >>> #
        >>> self.subdivide('region')
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist1 = ub.dict_hist([g['region'] for g in sampled])
        >>> print('hist1 = {}'.format(ub.urepr(hist1, nl=1)))
        >>> #
        >>> self.subdivide('color')
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist2 = ub.dict_hist([(g['region'],) + tuple(g['color'].keys()) for g in sampled])
        >>> print('hist2 = {}'.format(ub.urepr(hist2, nl=1)))

    Example:
        >>> # xdoctest: +REQUIRES(--benchmark)
        >>> from geowatch.tasks.fusion.datamodules.balanced_sampling import BalancedSampleForest
        >>> # Make a very large dataset to test speed constraints
        >>> sample_grid = [
        >>>     { 'region': 'region1', 'color': {'blue': 10, 'red': 3}},
        >>>     { 'region': 'region1', 'color': {'green': 3, 'purple': 2}},
        >>>     { 'region': 'region1', 'color': {'blue': 1}},
        >>>     { 'region': 'region1', 'color': {'green': 3, 'red': 5}},
        >>>     { 'region': 'region1', 'color': {'purple': 1, 'blue': 1}},
        >>>     { 'region': 'region2', 'color': {'blue': 5, 'red': 5}},
        >>>     { 'region': 'region2', 'color': {'green': 5, 'purple': 5}},
        >>> ] * 10000
        >>> #
        >>> self = BalancedSampleForest(sample_grid, rng=0)
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist0 = ub.dict_hist([g['region'] for g in sampled])
        >>> print('hist0 = {}'.format(ub.urepr(hist0, nl=1)))
        >>> #
        >>> self.subdivide('region')
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist1 = ub.dict_hist([g['region'] for g in sampled])
        >>> print('hist1 = {}'.format(ub.urepr(hist1, nl=1)))
        >>> #
        >>> self.subdivide('color')
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> hist2 = ub.dict_hist([(g['region'],) + tuple(g['color'].keys()) for g in sampled])
        >>> print('hist2 = {}'.format(ub.urepr(hist2, nl=1)))

    Example:
        >>> from geowatch.tasks.fusion.datamodules.balanced_sampling import BalancedSampleForest
        >>> from collections import Counter
        >>> # Imagine you have a dataset where each sample location could have 1 or
        >>> # more category. In this case make the value a dictionary where keys
        >>> # indicate the classes in the sample and the values are the importance
        >>> # of those classes to the sample.
        >>> # Make a very large dataset to test speed constraints
        >>> sample_grid = [
        >>>     { 'image_id': 1, 'class': {'dog': 1}},  # only 1 dog in this image
        >>>     { 'image_id': 2, 'class': {'dog': 1, 'cat': 2}},  # 1 dog and 2 cats in this image
        >>>     { 'image_id': 3, 'class': {'cat': 1}},
        >>>     { 'image_id': 4, 'class': {'cat': 1}},
        >>>     { 'image_id': 5, 'class': {'cat': 1}},
        >>>     { 'image_id': 6, 'class': {'cat': 1}},
        >>>     { 'image_id': 7, 'class': {'cat': 1}},
        >>>     { 'image_id': 8, 'class': {'cat': 1}},
        >>>     { 'image_id': 9, 'class': {'cat': 1}},
        >>>     { 'image_id': 10, 'class': {'cat': 3, 'dog': 1}}, # 3 cats and 1 dog in the image
        >>>     { 'image_id': 11, 'class': {'cat': 1, 'dog': 3}}, # 3 dogs and 1 cat in the image
        >>> ]
        >>> #
        >>> self = BalancedSampleForest(sample_grid, rng=0)
        >>> print(f'self={self}')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> class_counts = Counter()
        >>> for sample in sampled:
        >>>     class_counts.update(sample['class'])
        >>> print('Before Balancing')
        >>> print(f'class_counts = {ub.urepr(class_counts, nl=1)}')
        >>> # Do the balance step
        >>> self.subdivide('class')
        >>> sampled = list(ub.take(sample_grid, self._sample_many(100)))
        >>> class_counts = Counter()
        >>> for sample in sampled:
        >>>     class_counts.update(sample['class'])
        >>> print('After Balancing')
        >>> print(f'class_counts = {ub.urepr(class_counts, nl=1)}')

    TODO:
        Currently this will look at all attributes passed in each item in the
        sample grid. I think we want to specify what the attributes that could
        be balanced over are, which would help prevent a deep copy.
    """
    @profile
    def __init__(self, sample_grid, rng=None, n_trees=16, scoring='uniform'):
        """
        Args:
            sample_grid (List[Dict[str, Any]]) table of sample properties
            rng (int | None | RandomState):
                random number generator or seed
            n_trees (int): number of trees in the forest
            scoring (str): can be uniform or inverse
        """
        super().__init__()
        self.rng = rng = kwarray.ensure_rng(rng)
        self.n_trees = n_trees
        self.forest = self._create_forest(sample_grid, n_trees, scoring)

    def reseed(self, rng):
        """
        Reseed (or unseed) the random number generator

        Args:
            rng (int | None | RandomState):
                random number generator or seed
        """
        self.rng = kwarray.ensure_rng(rng)
        for tree in self.forest:
            tree.reseed(self.rng)

    @profile
    def _create_forest(self, sample_grid, n_trees, scoring):
        """
        Generate N BalancedSampleTree's, producing a hard assignment for
        each multi-label attribute. Expects a multi-label attribute to arrive
        as a dictionary with possible values as keys and frequencies as values.
        """
        import copy
        forest = []
        verbose = 1
        for idx in ub.ProgIter(range(n_trees), desc='Build balanced forests', verbose=verbose):
            local_sample_grid = copy.deepcopy(sample_grid)
            for sample in local_sample_grid:
                for key, val in sample.items():
                    if isinstance(val, dict):
                        if len(val) == 0:
                            sample[key] = None
                            continue
                        elif len(val) == 1:
                            sample[key] = list(val.keys())[0]
                            continue

                        # two or more choices
                        if scoring == 'inverse':
                            labels = list(val.keys())
                            freqs = np.asarray(list(val.values()))
                            weights = 1 - (freqs / freqs.sum())
                            weights = weights / weights.sum()
                            idx = self.rng.choice(len(labels), 1, p=weights)[0]
                            sample[key] = labels[idx]
                        elif scoring == 'uniform':
                            sample[key] = self.rng.choice(list(val.keys()))
                        else:
                            raise NotImplementedError(scoring)

            # initialize a BalancedSampleTree with this sample grid
            bst = BalancedSampleTree(local_sample_grid, rng=self.rng)
            forest.append(bst)
        return forest

    @profile
    def subdivide(self, key, weights=None, default_weight=0):
        """
        Adds a new layer to each tree in the forest that balances across the
        given attribute.

        Args:
            key (str):
                A key into the item dictionary of a sample that maps to the
                property to balance over.

            weights (None | Dict[Any, Number]):
                an optional mapping from values that ``key`` could point to
                to a numeric weight.

            default_weight (None | Number):
                if an attribute is unspecified in the weight table, this is
                the default weight it should be given. Default is 0.
        """
        for tree in self.forest:
            tree.subdivide(key, weights=weights, default_weight=default_weight)

    @profile
    def _sample_many(self, num):
        for _ in range(num):
            idx = self.sample()
            yield idx

    @profile
    def sample(self):
        """
        Uniformly sample a tree from the forest, then sample from it.

        Returns:
            int
        """
        idx = self.rng.choice(self.n_trees)
        return self.forest[idx].sample()

    @profile
    def __len__(self):
        return len(self.forest[0])

    @profile
    def __nice__(self):
        graph = self.forest[0].graph
        n_trees = self.n_trees
        n_nodes = graph.number_of_nodes()
        n_edges = graph.number_of_edges()
        n_leafs = len(self)
        n_depth = len(nx.algorithms.dag.dag_longest_path(graph))
        return f'trees={n_trees}, nodes={n_nodes}, edges={n_edges}, leafs={n_leafs}, depth={n_depth}'



"""
Module to define the :class:`BatchItem` wrapper class.

This is currently a work in progress, and the goal is to abstract the format of
the items produced by the datalaoder to make them ammenable for use with both
our heterogeneous networks as well as more standard networks that require data
be more regular.


This has a lot in common with prior work in:

    ~/code/netharn/netharn/data/data_containers.py
"""
import kwarray
import ubelt as ub
import numpy as np


class BatchItem(dict):
    """
    Ideally a batch item is simply an unstructured dictionary.
    This is the base class for more specific implementations, which are going
    to all be dictionaries, but the class will expose convinience methods.
    """

    def __repr__(self):
        """
        Returns:
            str
        """
        nice = self.__nice__()
        classname = self.__class__.__name__
        return '<{0}({1}) at {2}>'.format(classname, nice, hex(id(self)))

    def __str__(self):
        """
        Returns:
            str
        """
        classname = self.__class__.__name__
        nice = self.__nice__()
        return '<{0}({1})>'.format(classname, nice)

    def __nice__(self):
        return ''

    def asdict(self):
        return dict(self)

    def draw(self, **kwargs):
        raise NotImplementedError

    @classmethod
    def demo(cls, **kwargs):
        raise NotImplementedError

    @classmethod
    def summarize(cls, **kwargs):
        raise NotImplementedError


class HeterogeneousBatchItem(BatchItem):
    """
    A BatchItem is a container to help organize the output of the
    KWCocoVideoDataset. For backwards compatibility it retains the original
    dictionary interface.

    Example:
        >>> from geowatch.tasks.fusion.datamodules.network_io import *  # NOQA
        >>> self = HeterogeneousBatchItem.demo()
        >>> print(self)
        >>> print(ub.urepr(self.summarize(), nl=2))
        >>> canvas = self.draw()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas, fnum=1, pnum=(1, 1, 1))
        >>> kwplot.show_if_requested()
    """

    def __nice__(self):
        return f'num_frames={self.num_frames}, sensorchans={self.sensorchan_histogram}'

    @property
    def num_frames(self):
        return len(self['frames'])

    @property
    def sensorchan_histogram(self):
        histogram = ub.dict_hist(
            frame.get('sensor', '*') + ':' + mode_key
            for frame in self['frames']
            for mode_key in frame['modes'].keys()
        )
        return histogram

    @classmethod
    def demo(cls):
        from geowatch.tasks.fusion.datamodules import kwcoco_dataset
        import geowatch
        coco_dset = geowatch.coerce_kwcoco('geowatch-msi', num_frames=10)
        dataset = kwcoco_dataset.KWCocoVideoDataset(
            coco_dset, time_dims=4, window_dims=(300, 300),
            channels='auto')
        dataset.disable_augmenter = True
        index = dataset.sample_grid['targets'][dataset.sample_grid['positives_indexes'][0]]
        item = dataset[index]
        self = cls(**item)
        return self

    def draw(self, item_output=None, combinable_extra=None, max_channels=5,
             max_dim=224, norm_over_time='auto', overlay_on_image=False,
             draw_weights=True, rescale='auto', classes=None,
             predictable_classes=None, show_summary_text=True,
             requested_tasks=None, legend=True, **kwargs):
        """
        Visualize this batch item.
        Corresponds to the dataset :func:`IntrospectMixin.draw_item` method.

        Not finished. The dataset draw_item class has context not currently
        available like predictable classes, that needs to be represented in the
        item itself.
        """
        import kwimage
        item = self

        from geowatch import heuristics
        default_combinable_channels = [
            ub.oset(['red', 'green', 'blue']),
            ub.oset(['Dred', 'Dgreen', 'Dblue']),
            ub.oset(['r', 'g', 'b']),
            ub.oset(['impervious', 'forest', 'water']),
            ub.oset(['baren', 'field', 'water']),
            ub.oset(['landcover_hidden.0', 'landcover_hidden.1', 'landcover_hidden.2']),
            ub.oset(['sam.0', 'sam.1', 'sam.2']),
            ub.oset(['sam.3', 'sam.4', 'sam.5']),
        ] + heuristics.HUERISTIC_COMBINABLE_CHANNELS

        if requested_tasks is None:
            requested_tasks = {'class': True, 'saliency': True, 'change': True, 'outputs': False, 'boxes': False}
        # self.default_combinable_channels
        # if rescale == 'auto':
        #     rescale = self.config['input_space_scale'] != 'native'
        # if norm_over_time == 'auto':
        #     norm_over_time = self.config['normalize_peritem'] is not None

        # Hack to force the categories to draw right for SMART
        # FIXME: Use the correct class colors in visualization.
        requested_tasks = item['requested_tasks']
        predictable_classes = item['predictable_classes']
        if predictable_classes is not None:
            heuristics.ensure_heuristic_category_tree_colors(predictable_classes, force=True)

        from geowatch.tasks.fusion.datamodules.batch_visualization import BatchVisualizationBuilder
        builder = BatchVisualizationBuilder(
            item=item, item_output=item_output,
            default_combinable_channels=default_combinable_channels,
            norm_over_time=norm_over_time, max_dim=max_dim,
            max_channels=max_channels, overlay_on_image=overlay_on_image,
            draw_weights=draw_weights, combinable_extra=combinable_extra,
            classes=predictable_classes, requested_tasks=requested_tasks,
            rescale=rescale, **kwargs)
        canvas = builder.build()

        if show_summary_text:
            try:
                summary = item.summarize()
            except Exception as ex:
                summary = {'summary_error': str(ex)}
            summary = ub.udict(summary) - {'frame_summaries'}
            summary_text = ub.urepr(summary, nobr=1, precision=2, nl=-1)
            header = kwimage.draw_text_on_image(None, text=summary_text, halign='left', color='kitware_blue')
            canvas = kwimage.stack_images([canvas, header])

        if legend:
            from geowatch.tasks.fusion import utils
            label_to_color = {
                node: data['color']
                for node, data in self['predictable_classes'].graph.nodes.items()}
            legend_img = utils._memo_legend(label_to_color)
            legend_img = kwimage.imresize(legend_img, scale=4.0)
            canvas = kwimage.stack_images([canvas, legend_img], axis=1)

        return canvas

    def summarize(self, coco_dset=None, stats=False):
        """
        Return debugging stats about the item

        Args:
            coco_dset (CocoDataset):
                The coco dataset used to generate the item.
                If specified, allows the summary to lookup extra information

            stats (bool): if True, include statistics on input datas.

        Returns:
            dict : a summary of the item

        Example:
            >>> from geowatch.tasks.fusion.datamodules.network_io import *  # NOQA
            >>> self = HeterogeneousBatchItem.demo()
            >>> item_summary = self.summarize(stats=0)
            >>> print(f'item_summary = {ub.urepr(item_summary, nl=-2)}')
        """
        item = self
        item_summary = {}
        item_summary['frame_summaries'] = []
        timestamps = []
        for frame in item['frames']:
            frame_summary = {}
            for mode_key, im_mode in frame['modes'].items():
                domain_key = frame['sensor'] + ':' + mode_key
                frame_summary[domain_key] = {}
                if stats:
                    frame_summary[domain_key]['stats'] = kwarray.stats_dict(
                        im_mode, nan=True)
                frame_summary[domain_key]['shape'] = im_mode.shape
            label_keys = [
                'class_idxs', 'class_ohe', 'saliency', 'change'
                'class_weights', 'saliency_weights', 'change_weights',
                'output_weights', 'box_ltrb',
                # 'box_weights', 'box_tids', 'box_cidxs',
            ]
            for key in label_keys:
                if frame.get(key, None) is not None:
                    frame_summary[key] = {}
                    frame_summary[key]['shape'] = frame[key].shape
                    if stats:
                        frame_summary[key]['stats'] = kwarray.stats_dict(frame[key], nan=True)
            item_summary['frame_summaries'].append(frame_summary)
            if frame['date_captured']:
                timestamps.append(ub.timeparse(frame['date_captured']))

            if frame.get('ann_aids') is not None:
                if 0 and coco_dset is not None:
                    # disable as workaround for coco sql issues
                    # (i.e. we dont want to rely on a connection to the sql
                    # database in the main thread)
                    annots = coco_dset.annots(frame['ann_aids'])
                    cids = annots.lookup('category_id')
                    class_hist = ub.dict_hist(ub.udict(self.classes.id_to_node).take(cids))
                    frame_summary['class_hist'] = class_hist
                frame_summary['num_annots'] = len(frame['ann_aids'])

        vidname = item.get('video_name', None)
        if vidname is not None:
            item_summary['video_name'] = vidname
            if coco_dset is not None:
                try:
                    video = coco_dset.index.name_to_video[vidname]
                    vid_w = video['width']
                    vid_h = video['height']
                    item_summary['video_hw'] = (vid_h, vid_w)
                except (KeyError, AttributeError):
                    item_summary['video_hw'] = '?'

        if len(timestamps) > 1:
            deltas = np.diff(timestamps)
            deltas = [d.total_seconds() for d in deltas]
            item_summary['min_time'] = ub.timestamp(min(timestamps))
            item_summary['max_time'] = ub.timestamp(max(timestamps))
            if len(deltas):
                item_summary['min_delta'] = min(deltas)
                item_summary['max_delta'] = max(deltas)
                item_summary['mean_delta'] = np.mean(deltas)
        item_summary['input_gsd'] = item['input_gsd']
        item_summary['output_gsd'] = item['output_gsd']

        if 'requested_target' in item:
            item_summary['requested_target'] = item['requested_target']

        if 'target' in item:
            item_summary['resolved_target'] = item['target']

        item_summary['producer_rank'] = item.get('producer_rank', None)
        item_summary['producer_mode'] = item.get('producer_mode', None)
        item_summary['requested_index'] = item.get('requested_index', None)
        item_summary['resolved_index'] = item.get('resolved_index', None)
        return item_summary


class HomogeneousBatchItem(HeterogeneousBatchItem):
    """
    Ideally this is a simplified representation that "just works" with standard
    off the shelf networks.

    Example:
        >>> from geowatch.tasks.fusion.datamodules.network_io import *  # NOQA
        >>> self = HomogeneousBatchItem.demo()
        >>> print(self)
        >>> print(ub.urepr(self.summarize(), nl=2))
        >>> canvas = self.draw()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas, fnum=1, pnum=(1, 1, 1))
        >>> kwplot.show_if_requested()
    """

    @classmethod
    def demo(cls):
        """
        Example:
            cls = HomogeneousBatchItem
            self = cls.demo()
        """
        from geowatch.tasks.fusion.datamodules import kwcoco_dataset
        import geowatch
        coco_dset = geowatch.coerce_kwcoco('vidshapes1', num_frames=10)
        dataset = kwcoco_dataset.KWCocoVideoDataset(
            coco_dset, time_dims=3, window_dims=(300, 300),
            channels='r|g|b')
        dataset.disable_augmenter = True
        index = dataset.sample_grid['targets'][dataset.sample_grid['positives_indexes'][0]]
        item = dataset[index]
        self = cls(**item)
        return self


class RGBImageBatchItem(HomogeneousBatchItem):
    """
    Only allows a single RGB image as the input.

    Example:
        >>> from geowatch.tasks.fusion.datamodules.network_io import *  # NOQA
        >>> self = RGBImageBatchItem.demo()
        >>> print(self)
        >>> print(ub.urepr(self.summarize(), nl=2))
        >>> canvas = self.draw()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas, fnum=1, pnum=(1, 1, 1))
        >>> kwplot.show_if_requested()
    """
    def __nice__(self):
        return f'num_frames={self.num_frames}, sensorchans={self.sensorchan_histogram}'

    @property
    def frame(self):
        frame = self['frames'][0]
        return frame

    @property
    def channels(self):
        frame = self.frame
        modes = frame['modes']
        mode_keys = list(modes.keys())
        assert len(mode_keys) == 1
        mode_key = mode_keys[0]
        return mode_key

    @property
    def imdata_chw(self):
        frame = self.frame
        modes = frame['modes']
        mode_keys = list(modes.keys())
        assert len(mode_keys) == 1
        mode_key = mode_keys[0]
        mode_val = modes[mode_key]
        imdata_chw = mode_val
        return imdata_chw

    @property
    def nonlocal_class_ohe(self):
        frame = self.frame
        nonlocal_class_ohe = frame['nonlocal_class_ohe']
        return nonlocal_class_ohe

    @classmethod
    def demo(cls, index=None):
        """
        cls = RGBImageBatchItem
        """
        dataset = _demo_dataset()
        if index is None:
            index = dataset.sample_grid['targets'][dataset.sample_grid['positives_indexes'][0]]
        item = dataset[index]
        self = cls(**item)
        assert len(item['frames']) == 1
        frame = self['frames'][0]
        modes = frame['modes']
        mode_keys = list(modes.keys())
        assert len(mode_keys) == 1
        # mode_key = mode_keys[0]
        # mode_val = modes[mode_key]
        # new_item = ub.udict(item) - {'frames'}
        # new_frame = ub.udict(frame) - {'modes'}
        # new_frame['channels'] = mode_key
        # new_frame['imdata_chw'] = mode_val
        # new_item.update(new_frame)
        return self


@ub.memoize
def _demo_dataset():
    """
    Cached dataset for more efficient testing
    """
    from geowatch.tasks.fusion.datamodules import kwcoco_dataset
    import geowatch
    coco_dset = geowatch.coerce_kwcoco('vidshapes1', num_frames=1)
    dataset = kwcoco_dataset.KWCocoVideoDataset(
        coco_dset, time_dims=1, window_dims=(300, 300),
        channels='r|g|b')

    dataset.requested_tasks['nonlocal_class'] = True
    dataset.disable_augmenter = True
    return dataset


# ----------------------------------
# Non-collated Batch Item Containers
# ----------------------------------


class UncollatedBatch(list):
    """
    A generic list of batch items, which may or may not be collatable.
    """


class HeterogeneousBatch(UncollatedBatch):
    """
    A HeterogeneousBatch a ``List[HeterogeneousBatchItem]``.
    """


class UncollatedRGBImageBatch(UncollatedBatch):
    """
    A list of collatable RGBImageBatchItem
    """

    @classmethod
    def demo(cls, num_items=3):
        """
        """
        self = cls.from_items(RGBImageBatchItem.demo(_) for _ in range(num_items))
        return self

    @classmethod
    def coerce(cls, data):
        if isinstance(data, list):
            self = cls.from_items(data)
        else:
            raise NotImplementedError
        return self

    @classmethod
    def from_items(cls, data):
        self = cls(RGBImageBatchItem(item) for item in data)
        return self

    def collate(self):
        """
        Returns:
            CollatedRGBImageBatch

        Example:
            >>> from geowatch.tasks.fusion.datamodules.network_io import *  # NOQA
            >>> self = UncollatedRGBImageBatch.demo()
            >>> batch = self.collate()
        """
        import torch
        imdatas = [batch_item.imdata_chw for batch_item in self]
        imdata_bchw = torch.stack(imdatas)

        batch = CollatedRGBImageBatch()
        batch['imdata_bchw'] = imdata_bchw

        try:
            nonlocal_class_ohes = [batch_item.nonlocal_class_ohe for batch_item in self]
        except KeyError:
            ...
        else:
            nonlocal_class_ohe = torch.stack(nonlocal_class_ohes)
            batch['nonlocal_class_ohe'] = nonlocal_class_ohe

        return batch


# ------------------------------
# Collated Batch Item Containers
# ------------------------------


class CollatedBatch(dict):

    def asdict(self):
        return dict(self)


class CollatedRGBImageBatch(CollatedBatch):

    def to(self, device):
        for k, v in self.items():
            self[k] = v.to(device)
        return self


# ---------------
# Network Outputs
# ---------------
# The concept of a network input needs to be mirrored by a network output.


class NetworkOutputs(dict):
    """
    Network outputs should ALWAYS be a dictionary, this is the most flexible
    way to encode networks such that they can be extended later.
    """

    def _debug_shape(self):
        from geowatch.utils.util_netharn import _debug_inbatch_shapes
        _debug_inbatch_shapes(self)


# ------------------------------------
# Collated Network Output Containers
# ------------------------------------

class UncollatedNetworkOutputs(NetworkOutputs):
    ...


# ------------------------------------
# Uncollated Network Output Containers
# ------------------------------------

class CollatedNetworkOutputs(NetworkOutputs):
    """
    Example:
        >>> from geowatch.tasks.fusion.datamodules.network_io import *  # NOQA
        >>> B, H, W, C = 2, 3, 3, 11
        >>> import torch
        >>> logits = {
        >>>     'nonlocal_class': (torch.rand(B, C) - 0.5) * 10,
        >>>     'segmentation_class': (torch.rand(B, W, H, C) - 0.5) * 10,
        >>>     'nonlocal_saliency': (torch.rand(B, 1) - 0.5) * 10,
        >>>     'segmentation_saliency': (torch.rand(B, W, H, 1) - 0.5) * 10,
        >>> }
        >>> self = CollatedNetworkOutputs(
        >>>     logits=logits,
        >>>     probs={k + '_probs': v.sigmoid() for k, v in logits.items()},
        >>>     loss_parts={},
        >>>     loss=10,
        >>> )
        >>> new = self.decollate()
        >>> self._debug_shape()
        >>> new._debug_shape()
    """

    def decollate(self):
        """
        Convert back into a per-item structure for easier analysis / drawing.
        """
        new = UncollatedNetworkOutputs()
        new['loss'] = self.get('loss', None)
        new['item_probs'] = decollate(self['probs'])
        return new


def decollate(collated):
    """
    Breakup a collated batch in a standardized way.
    Returns a list of items for each batch item with a structure that matches
    the collated batch, but without the leading batch dimension in each value.

    Example:
        >>> from geowatch.tasks.fusion.datamodules.network_io import *  # NOQA
        >>> import torch
        >>> B, H, W, C = 5, 2, 3, 7
        >>> collated = {
        >>>     'segmentation_class': torch.rand(B, H, W, C),
        >>>     'nonlocal_class': torch.rand(B, C),
        >>> }
        >>> uncollated = decollate(collated)
        >>> assert len(uncollated) == B
        >>> assert (uncollated[0]['nonlocal_class'] == collated['nonlocal_class'][0]).all()
    """
    import ubelt as ub
    # TODO: make more efficient?
    walker = ub.IndexableWalker(collated)
    uncollated_dict = ub.AutoDict()
    uncollated_walker = ub.IndexableWalker(uncollated_dict)
    for path, batch_val in walker:
        for bx, item_val in enumerate(batch_val):
            uncollated_walker[[bx] + path] = item_val
    uncollated = list(uncollated_dict.to_dict().values())
    return uncollated



"""
Describe how to interpret QA bands.

References:
    https://smart-research.slack.com/?redir=%2Ffiles%2FU028UQGN1N0%2FF04B998ANRL%2Faccenture_ta1_productdoc_phaseii_20211117.pptx%3Forigin_team%3DTN3QR7WAH%26origin_channel%3DC03QTAXU7GF
"""
import ubelt as ub
import functools
import operator
import numpy as np
import math
from kwutil import util_pattern


def _dump_qa_debug_vid():
    """
    Make human interpretable sequences of QA bands and RGB data.
    """
    import kwcoco
    import geowatch
    import kwimage
    data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
    coco_fpath = data_dvc_dpath / 'Drop4-BAS/KR_R001.kwcoco.json'
    dset = kwcoco.CocoDataset(coco_fpath)

    videos = dset.videos()
    images = videos.images[0]

    dump_dpath = ub.Path('_dump_qa_debug_vid').ensuredir()

    for idx, gid in enumerate(images):
        coco_img = dset.coco_image(gid)
        # Load some quality and rgb data
        qa_delayed = coco_img.imdelay('cloudmask', interpolation='nearest', antialias=False)
        rgb_delayed = coco_img.imdelay('red|green|blue')
        quality_im = qa_delayed.finalize()
        rgb_canvas = kwimage.normalize_intensity(rgb_delayed.finalize(nodata_method='float'))
        sensor = coco_img.img.get('sensor_coarse')
        print(f'sensor={sensor}')
        # Use the spec to draw it
        from geowatch.tasks.fusion.datamodules.qa_bands import QA_SPECS
        table = QA_SPECS.find_table('ACC-1', 'L8')
        #table = QA_SPECS.find_table('FMASK', '*')
        #table = QA_SPECS.find_table('Phase1_QA', '*')
        drawings = table.draw_labels(quality_im)
        qa_canvas = drawings['qa_canvas']
        legend = drawings['legend']
        canvas = kwimage.stack_images([rgb_canvas, qa_canvas, legend], axis=1)

        spec_name = table.spec['qa_spec_name']
        canvas = kwimage.draw_header_text(canvas, sensor + ' ' + spec_name)

        _kw = ub.compatible({'on_value': 0.3}, kwimage.fill_nans_with_checkers)
        canvas = kwimage.fill_nans_with_checkers(canvas, **_kw)

        fname = f'frame_{idx:08d}.jpg'
        fpath = dump_dpath / fname
        kwimage.imwrite(fpath, kwimage.ensure_uint255(canvas))

    #canvas = kwimage.stack_images([canvas, ], axis=0)
    import kwplot
    kwplot.autoplt()
    kwplot.imshow(canvas)


class QA_SpecMixin:

    def draw_labels(table, quality_im, legend='separate', legend_dpi=96, verbose=0):
        """

        The doctest can be used to debug cloudmasks for the datasets

        Ignore:
            >>> import kwcoco
            >>> import geowatch
            >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
            >>> dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data')
            >>> coco_fpath = dvc_dpath / 'Drop6/data_vali_split1.kwcoco.zip'
            >>> dset = kwcoco.CocoDataset(coco_fpath)
            >>> gid = dset.images()[18]
            >>> coco_img = dset.coco_image(gid)
            >>> # Load some quality and rgb data
            >>> qa_delayed = coco_img.imdelay('cloudmask', interpolation='nearest', antialias=False)
            >>> rgb_delayed = coco_img.imdelay('red|green|blue')
            >>> quality_im = qa_delayed.finalize()
            >>> rgb_canvas = kwimage.normalize_intensity(rgb_delayed.finalize(nodata_method='float'))
            >>> sensor = coco_img.img.get('sensor_coarse')
            >>> print(f'sensor={sensor}')
            >>> # Use the spec to draw it
            >>> from geowatch.tasks.fusion.datamodules.qa_bands import QA_SPECS
            >>> table = QA_SPECS.find_table('ACC-1', 'L8')
            >>> #table = QA_SPECS.find_table('FMASK', '*')
            >>> #table = QA_SPECS.find_table('Phase1_QA', '*')
            >>> drawings = table.draw_labels(quality_im)
            >>> qa_canvas = drawings['qa_canvas']
            >>> legend = drawings['legend']
            >>> canvas = kwimage.stack_images([rgb_canvas, qa_canvas, legend], axis=1)
            >>> canvas = kwimage.draw_header_text(canvas, sensor)
            >>> #canvas = kwimage.stack_images([canvas, ], axis=0)
            >>> import kwplot
            >>> kwplot.autoplt()
            >>> kwplot.imshow(canvas)

        Ignore:
            >>> import kwcoco
            >>> import geowatch
            >>> import kwimage
            >>> data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='hdd')
            >>> dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data')
            >>> coco_fpath = dvc_dpath / 'Aligned-Drop7/KR_R002/imgonly-KR_R002.kwcoco.zip'
            >>> dset = kwcoco.CocoDataset(coco_fpath)
            >>> wv_gids = [g for g, s in dset.images().lookup('sensor_coarse', keepid=True).items() if s == 'WV']
            >>> gid = dset.images(wv_gids)[-3]
            >>> coco_img = dset.coco_image(gid)
            >>> # Load some quality and rgb data
            >>> qa_delayed = coco_img.imdelay('quality', interpolation='nearest', antialias=False)
            >>> rgb_delayed = coco_img.imdelay('red|green|blue')
            >>> quality_im = qa_delayed.finalize(interpolation='nearest', antialias=False)
            >>> rgb_raw = rgb_delayed.finalize(nodata_method='float')
            >>> rgb_canvas = kwimage.normalize_intensity(rgb_raw)
            >>> sensor = coco_img.img.get('sensor_coarse')
            >>> print(f'sensor={sensor}')
            >>> # Use the spec to draw it
            >>> from geowatch.tasks.fusion.datamodules.qa_bands import QA_SPECS
            >>> table = QA_SPECS.find_table('ACC-1', 'WV')
            >>> #table = QA_SPECS.find_table('FMASK', '*')
            >>> #table = QA_SPECS.find_table('Phase1_QA', '*')
            >>> drawings = table.draw_labels(quality_im)
            >>> qa_canvas = drawings['qa_canvas']
            >>> legend = drawings['legend']
            >>> canvas = kwimage.stack_images([rgb_canvas, qa_canvas], axis=1)
            >>> canvas = kwimage.draw_header_text(canvas, sensor)
            >>> #canvas = kwimage.stack_images([canvas, ], axis=0)
            >>> import kwplot
            >>> kwplot.autoplt()
            >>> kwplot.imshow(canvas, fnum=1)
            >>> kwplot.imshow(legend, fnum=2)
        """
        import kwimage
        import kwarray
        import numpy as np
        import kwplot

        if verbose:
            print(f'Build quality image for {quality_im.shape}')

        _raw = quality_im.ravel()

        if quality_im.dtype.kind not in {'u', 'i'}:
            if verbose:
                print('Check for nan')
            is_nan = np.isnan(_raw)
            num_nan = is_nan.sum()
            _raw2 = _raw[~is_nan]
        else:
            if hasattr(quality_im, 'mask'):
                is_nan = quality_im.mask
                num_nan = is_nan.sum()
                _raw2 = quality_im.data[~is_nan].ravel()
            else:
                is_nan = None
                num_nan = 0
                _raw2 = _raw

        if verbose:
            print('Counting unique values')

        # qavals_to_count = ub.dict_hist(_raw2)
        qavals, counts = np.unique(_raw2, return_counts=True)
        qavals_to_count = ub.dzip(qavals, counts)

        if num_nan:
            print('warning nan QA')
            # qavals_to_count[np.nan] = num_nan

        unique_qavals = list(qavals_to_count.keys())

        if verbose:
            print(f'Found {len(unique_qavals)} unique labels')

        max_labels = 32
        if len(qavals_to_count) > max_labels:
            print('WARNING: QA band has a lot of unique values')
            top_qvals = dict(list(ub.udict(qavals_to_count).sorted_values().items())[-max_labels:])
            unique_qavals = list(top_qvals)
            # qval_to_color = ub.udict(qval_to_color)
            # qval_to_color = qval_to_color.subdict(top_qvals)

        # For the QA band lets assign a color to each category
        colors = kwimage.Color.distinct(len(unique_qavals))
        qval_to_color = dict(zip(unique_qavals, colors))

        qval_to_desc = table.describe_values(unique_qavals)
        quality_im = kwarray.atleast_nd(quality_im, 3)

        # Colorize the QA bands
        if verbose:
            print('Colorizing')

        colorized = np.empty(quality_im.shape[0:2] + (3,), dtype=np.uint8)
        if len(qval_to_color) > 10:
            qa_iter = ub.ProgIter(qval_to_color.items(), total=len(qval_to_color), desc='complex QA')
        else:
            qa_iter = qval_to_color.items()
        for qabit, color in ub.ProgIter(qa_iter, desc='colorize', enabled=verbose):
            color255 = kwimage.Color.coerce(color).as255()
            mask = quality_im[:, :, 0] == qabit
            colorized[mask] = color255

        # Because the QA band is categorical, we should be able to make a short
        if is_nan is not None:
            masked = np.ma.MaskedArray(colorized, mask=np.tile(is_nan[..., None], (1, 1, 3)))
            qa_canvas = kwimage.nodata_checkerboard(masked).data
        else:
            qa_canvas = colorized

        label_to_color = ub.udict(qval_to_color).map_keys(qval_to_desc.__getitem__)

        if verbose:
            print('Build legend')
        legend = kwplot.make_legend_img(label_to_color, dpi=legend_dpi)  # Make a legend

        if verbose:
            print('finished qa drawing')

        drawings = {
            'qa_canvas': qa_canvas,
            'legend': legend,
        }
        return drawings


class QA_BitSpecTable(QA_SpecMixin):
    """
    Bit tables are more efficient because we can reduce over the query input

    Example:
        >>> from geowatch.tasks.fusion.datamodules import qa_bands
        >>> import kwimage
        >>> # Lookup a table for this spec
        >>> self = qa_bands.QA_SPECS.find_table('ACC-1', 'S2')
        >>> assert isinstance(self, qa_bands.QA_BitSpecTable)
        >>> # Make a quality image with every value
        >>> pure_patches = [np.zeros((32, 32), dtype=np.int16) + val for val in self.name_to_value.values()]
        >>> # Also add in a few mixed patches
        >>> mixed_patches = [
        >>>     pure_patches[0] | pure_patches[1],
        >>>     pure_patches[4] | pure_patches[1],
        >>>     pure_patches[3] | pure_patches[5],
        >>>     pure_patches[0] | pure_patches[4],
        >>>     pure_patches[3] | pure_patches[4],
        >>> ]
        >>> patches = pure_patches + mixed_patches
        >>> quality_im = kwimage.stack_images_grid(patches)
        >>> # The mask_any method makes a mask where any of the semantically given labels will be masked
        >>> query_names = ['cloud']
        >>> is_iffy = self.mask_any(quality_im, ['cloud', 'cirrus'])
        >>> drawings = self.draw_labels(quality_im)  # visualize
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> qa_canvas = drawings['qa_canvas']
        >>> legend = drawings['legend']
        >>> kwplot.imshow(is_iffy, pnum=(1, 3, 1), title=f'mask matching {query_names}')
        >>> kwplot.imshow(qa_canvas, pnum=(1, 3, 2), title='qa bits')
        >>> kwplot.imshow(legend, pnum=(1, 3, 3), title='qa bit legend')
        >>> kwplot.set_figtitle(f"QA Spec: name={self.spec['qa_spec_name']} sensor={self.spec['sensor']}")

    Example:
        >>> from geowatch.tasks.fusion.datamodules import qa_bands
        >>> import kwimage
        >>> # Lookup a table for this spec
        >>> self = qa_bands.QA_SPECS.find_table('qa_pixel', 'L8')
        >>> assert isinstance(self, qa_bands.QA_BitSpecTable)
        >>> # Make a quality image with every value
        >>> pure_patches = [np.zeros((32, 32), dtype=np.int16) + val for val in self.name_to_value.values()]
        >>> # Also add in a few mixed patches
        >>> mixed_patches = [
        >>>     pure_patches[0] | pure_patches[1],
        >>>     pure_patches[2] | pure_patches[1],
        >>> ]
        >>> patches = pure_patches + mixed_patches
        >>> quality_im = kwimage.stack_images_grid(patches)
        >>> # The mask_any method makes a mask where any of the semantically given labels will be masked
        >>> query_names = ['cloud']
        >>> is_iffy = self.mask_any(quality_im, ['cloud'])
        >>> drawings = self.draw_labels(quality_im)  # visualize
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> qa_canvas = drawings['qa_canvas']
        >>> legend = drawings['legend']
        >>> kwplot.imshow(is_iffy, pnum=(1, 3, 1), title=f'mask matching {query_names}')
        >>> kwplot.imshow(qa_canvas, pnum=(1, 3, 2), title='qa bits')
        >>> kwplot.imshow(legend, pnum=(1, 3, 3), title='qa bit legend')
        >>> kwplot.set_figtitle(f"QA Spec: name={self.spec['qa_spec_name']} sensor={self.spec['sensor']}")
    """
    def __init__(table, spec):
        table.spec = spec
        table.name_to_value = ub.udict({
            item['qa_name']: 1 << item['bit_number']
            for item in table.spec['bits']
            if item.get('qa_name', None) is not None
        })

        ranged_bits = []
        for item in table.spec.get('bit_ranges', []):
            if item.get('qa_name', None) is not None:
                ranged_bits.extend(item['bit_range'])
                # print('item = {}'.format(ub.urepr(item, nl=1)))
        table.ranged_bits = ranged_bits

    def mask_any(table, quality_im, qa_names):
        if quality_im.dtype.kind == 'f':
            raise ValueError('The quality mask should be an bitwise integer type, not a float')
        bit_values = list((table.name_to_value & qa_names).values())
        iffy_bits = functools.reduce(operator.or_, bit_values)
        is_iffy = (quality_im & iffy_bits) > 0
        return is_iffy

    def describe_values(table, unique_qavals):
        """
        Get a human readable description of each value for a legend
        """
        bit_to_spec = {}
        for item in table.spec['bits']:
            bit_to_spec[item['bit_number']] = item

        val_to_desc = {}

        for val in unique_qavals:
            # For each value determine what bits are on
            if val >= 0:
                bit_positions = unpack_bit_positions(val)

                descs = []
                for bit_number in bit_positions:
                    if bit_number in table.ranged_bits:
                        continue
                    bit_spec = bit_to_spec.get(bit_number, '?')
                    if bit_spec == '?':
                        descs.append('?')
                    else:
                        descs.append(bit_spec['qa_description'])

                parts = {}
                parts['value'] = val
                parts['bits'] = '|'.join(list(map(str, bit_positions)))
                parts['desc'] = ',\n'.join(descs)
            else:
                parts = {}
                parts['value'] = val
                parts['bits'] = '---'
                parts['desc'] = 'nodata'
            val_to_desc[val] = '───\n' + ub.urepr(parts, compact=1, nobr=1, nl=True, si=1, sort=0) + '\n───'
        return val_to_desc


def unpack_bit_positions(val, itemsize=None):
    """
    Given an integer value, return the positions of the on bits.

    Args:
        val (int): a signed or unsigned integer

        itemsize (int | None):
            Number of bytes used to represent the integer. E.g. 1 for a uint8 4
            for an int32. If unspecified infer the smallest number of bytes
            needed, but warning this may produce ambiguous results for negative
            numbers.

    Returns:
        List[int]: the indexes of the 1 bits.

    Note:
        This turns out to be faster than a numpy or lookuptable strategy I
        tried.  See github.com:Erotemic/misc/learn/bit_conversion.py

    Example:
        >>> unpack_bit_positions(0)
        []
        >>> unpack_bit_positions(1)
        [0]
        >>> unpack_bit_positions(-1)
        [0, 1, 2, 3, 4, 5, 6, 7]
        >>> unpack_bit_positions(-1, itemsize=2)
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
        >>> unpack_bit_positions(9)
        [0, 3]
        >>> unpack_bit_positions(2132)
        [2, 4, 6, 11]
        >>> unpack_bit_positions(-9999)
        [0, 4, 5, 6, 7, 11, 12, 14, 15]
        >>> unpack_bit_positions(np.int16(-9999))
        [0, 4, 5, 6, 7, 11, 12, 14, 15]
        >>> unpack_bit_positions(np.int16(-1))
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
    """
    is_negative = val < 0
    if is_negative:
        if itemsize is None:
            try:
                bit_length = val.bit_length() + 1
                itemsize = math.ceil(bit_length / 8.0)  # bytelength
            except AttributeError:
                # Probably a numpy type
                itemsize = val.dtype.itemsize
        neg_position = (itemsize * 8) - 1
        # special logic for negatives to get twos compliment repr
        max_val = 1 << neg_position
        try:
            val_ = max_val + val
        except OverflowError:
            # numpy 2.x
            val_ = max_val + int(val)
    else:
        val_ = val
    binary_string = '{:b}'.format(val_)[::-1]
    bit_positions = [pos for pos, char in enumerate(binary_string)
                     if char == '1']
    if is_negative:
        bit_positions.append(neg_position)
    return bit_positions


class QA_ValueSpecTable(QA_SpecMixin):
    """
    Value tables are less efficient
    """
    def __init__(table, spec):
        table.spec = spec
        table.name_to_value = ub.udict({
            item['qa_name']: item['value']
            for item in table.spec['values']
            if item.get('qa_name', None) is not None
        })

    def mask_any(table, quality_im, qa_names):
        if quality_im.dtype.kind == 'f':
            raise ValueError('The quality mask should be an integer type, not a float')
        iffy_values = list((table.name_to_value & qa_names).values())
        is_iffy = np.logical_or.reduce([quality_im == value for value in iffy_values])
        return is_iffy

    def describe_values(table, unique_qavals):
        """
        Get a human readable description of each value for a legend
        """

        val_to_spec = {}
        for item in table.spec['values']:
            val_to_spec[item['value']] = item
        val_to_desc = {}
        for val in unique_qavals:
            # For each value determine what bits are on
            spec = val_to_spec.get(val, None)
            parts = {}
            parts['value'] = val
            if spec is None:
                val_to_desc[val] = '?'
            else:
                parts['desc'] = spec['qa_description']
            val_to_desc[val] = ub.urepr(parts, compact=1, nobr=1, nl=True, si=1, sort=0)
        return val_to_desc


class QA_SpecRegistry(list):

    def query_table(self, spec_name='*', sensor='*'):
        """
        Ignore:
            from geowatch.tasks.fusion.datamodules.qa_bands import *  # NOQA
            self = QA_SPECS
            spec_name = 'ACC-1'
            sensor = 'L8'
            table, = list(self.query_table(spec_name, sensor))
            qa_names = ['cloud']
            table.mask_for_any(qa_names)
        """
        spec_pat = util_pattern.Pattern.coerce(spec_name)
        sensor_pat = util_pattern.Pattern.coerce(sensor)
        for table in self:
            matches_main_sensor = sensor_pat.match(table.spec['sensor'])
            maches_sensor_alias = any(
                sensor_pat.match(a) for a in table.spec.get('sensor_alias', []))
            matches_spec_name = spec_pat.match(table.spec['qa_spec_name'])
            matches_sensor = matches_main_sensor or maches_sensor_alias
            if matches_spec_name and matches_sensor:
                yield table

    def find_table(self, spec_name='*', sensor='*'):
        results = list(self.query_table(spec_name=spec_name, sensor=sensor))
        if len(results) != 1:
            raise AssertionError(f'{len(results)} - {spec_name}, {sensor}')
        table = results[0]
        return table


QA_SPECS = QA_SpecRegistry()


QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'ACC-1',
    'qa_spec_date': '2022-11-28',
    'sensor': 'S2',
    'sensor_alias': ['Sentinel-2'],
    'dtype': {'kind': 'u', 'itemsize': 2},
    'bits': [
        {
            'bit_number': 0,
            'qa_name': 'combined',
            'qa_description': 'combined qa mask',
            'bit_value': [
                {'value': 1, 'description': 'use-pixel'}
            ]
        },
        {
            'bit_number': 1,
            'qa_name': 'cloud',
            'qa_description': 'cloud',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 2,
            'qa_name': 'cloud_adjacent',
            'qa_description': 'adjacent to cloud/shadow',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 3,
            'qa_name': 'cloud_shadow',
            'qa_description': 'cloud shadow',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 4,
            'qa_name': 'ice',
            'qa_description': 'snow / ice',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 5,
            'qa_name': 'water',
            'qa_description': 'water',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {'bit_number': 6, 'qa_name': None, 'qa_description': 'reserved for future use'},
        {'bit_number': 7, 'qa_name': None, 'qa_description': 'reserved for future use'},
        {
            'bit_number': 8,
            'qa_name': 'filled',
            'qa_description': 'filled value',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {'bit_number': 9, 'qa_description': 'reserved for future use'},
        {'bit_number': 10, 'qa_description': 'reserved for future use'},
        {'bit_number': 11, 'qa_description': 'reserved for future use'},
        {'bit_number': 12, 'qa_description': 'reserved for future use'},
        {'bit_number': 13, 'qa_description': 'reserved for future use'},
        {'bit_number': 14, 'qa_description': 'reserved for future use'},
        {'bit_number': 15, 'qa_description': 'reserved for future use'}
    ]
}))


QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'ACC-1',
    'qa_spec_date': '2022-11-28',
    'sensor': 'L8',
    'sensor_alias': ['Landsat 8'],
    'dtype': {'kind': 'u', 'itemsize': 2},
    'bits': [
        {
            'bit_number': 0,
            'qa_name': 'combined',
            'qa_description': 'combined qa mask',
            'bit_value': [
                {'value': 1, 'description': 'use-pixel'}
            ]
        },
        {
            'bit_number': 1,
            'qa_name': 'cloud',
            'qa_description': 'cloud',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 2,
            'qa_name': 'cloud_adjacent',
            'qa_description': 'adjacent to cloud/shadow',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 3,
            'qa_name': 'cloud_shadow',
            'qa_description': 'cloud shadow',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 4,
            'qa_name': 'ice',
            'qa_description': 'snow / ice',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 5,
            'qa_name': 'water',
            'qa_description': 'water',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {'bit_number': 6, 'qa_name': None, 'qa_description': 'reserved for future use'},
        {'bit_number': 7, 'qa_name': None, 'qa_description': 'reserved for future use'},
        {
            'bit_number': 8,
            'qa_name': 'filled',
            'qa_description': 'filled value',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {'bit_number': 9, 'qa_description': 'reserved for future use'},
        {'bit_number': 10, 'qa_description': 'reserved for future use'},
        {'bit_number': 11, 'qa_description': 'reserved for future use'},
        {'bit_number': 12, 'qa_description': 'reserved for future use'},
        {'bit_number': 13, 'qa_description': 'reserved for future use'},
        {'bit_number': 14, 'qa_description': 'reserved for future use'},
        {'bit_number': 15, 'qa_description': 'reserved for future use'}
    ]
}))


"""
Note: In Ver 2 processing, WorldView VNIR products will include a 2nd QA file
containing QA information per multispectral band -Filename: *_QA2.tif -Format
same as original QA file
"""
QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'ACC-1',
    'qa_spec_date': '2022-11-28',
    'sensor': 'WV',
    'sensor_alias': ['WV1'],
    'dtype': {'kind': 'u', 'itemsize': 2},
    'bits': [
        {
            'bit_number': 0,
            'qa_name': 'combined',
            'qa_description': 'combined qa mask',
            'bit_value': [
                {'value': 1, 'description': 'use-pixel'},
                {'value': 0, 'description': 'ignore-pixel'}
            ]
        },
        {
            'bit_number': 1,
            'qa_name': 'cloud',
            'qa_description': 'cloud',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 2,
            'qa_name': 'cloud_shadow',
            'qa_description': 'cloud shadow',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 3,
            'qa_name': 'thin_cloud',
            'qa_description': 'thin cloud',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {'bit_number': 4, 'qa_description': 'reserved for future use'},
        {'bit_number': 7, 'qa_description': 'reserved for future use'},
        {
            'bit_number': 8,
            'qa_description': 'filled value / suspicious pixel',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 9,
            'qa_description': 'AOD Source',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 10,
            'qa_description': 'climatology source',
            'bit_value': [
                {'value': 1, 'description': 'climatology'},
                {'value': 0, 'description': 'MODIS'}
            ]
        },
        {'bit_number': 11, 'qa_description': 'reserved for future use'},
        {'bit_number': 12, 'qa_description': 'reserved for future use'},
        {'bit_number': 13, 'qa_description': 'reserved for future use'},
        {'bit_number': 14, 'qa_description': 'reserved for future use'},
        {'bit_number': 15, 'qa_description': 'reserved for future use'}
    ]
}))


QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'ACC-1',
    'qa_spec_date': '2022-11-28',
    'sensor': 'PD',
    'dtype': {'kind': 'u', 'itemsize': 2},
    'bits': [
        {
            'bit_number': 0,
            'qa_name': 'combined',
            'qa_description': 'combined qa mask',
            'bit_value': [
                {'value': 1, 'description': 'use-pixel'},
                {'value': 0, 'description': 'ignore-pixel'}
            ]
        },
        {
            'bit_number': 1,
            'qa_name': 'cloud',
            'qa_description': 'cloud',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {'bit_number': 2, 'qa_description': 'reserved for future use'},
        {'bit_number': 7, 'qa_description': 'reserved for future use'},
        {
            'bit_number': 8,
            'qa_name': 'filled',
            'qa_description': 'filled value / suspicious pixel',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 9,
            'qa_description': 'AOD Source',
            'bit_value': [
                {'value': 1, 'description': 'yes'},
                {'value': 0, 'description': 'no'}
            ]
        },
        {
            'bit_number': 10,
            'qa_description': 'climatology source',
            'bit_value': [
                {'value': 1, 'description': 'climatology'},
                {'value': 0, 'description': 'MODIS'}
            ]
        },
        {'bit_number': 11, 'qa_description': 'reserved for future use'},
        {'bit_number': 12, 'qa_description': 'reserved for future use'},
        {'bit_number': 13, 'qa_description': 'reserved for future use'},
        {'bit_number': 14, 'qa_description': 'reserved for future use'},
        {'bit_number': 15, 'qa_description': 'reserved for future use'}
    ]
}))


# https://github.com/GERSL/Fmask#46-version
QA_SPECS.append(QA_ValueSpecTable({
    'qa_spec_name': 'FMASK',
    'qa_spec_date': '???',
    'sensor': '*',
    'dtype': {'kind': 'u', 'itemsize': 1},
    'values': [
        {'value': 0, 'qa_name': 'clear', 'qa_description': 'clear land pixel'},
        {'value': 1, 'qa_name': 'water', 'qa_description': 'clear water pixel'},
        {'value': 2, 'qa_name': 'cloud_shadow', 'qa_description': 'cloud shadow'},
        {'value': 3, 'qa_name': 'ice', 'qa_description': 'snow'},
        {'value': 4, 'qa_name': 'cloud', 'qa_description': 'cloud'},
        {'value': 255, 'qa_description': 'no observation'}
    ]
}))


# https://github.com/GERSL/Fmask#46-version
QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'Phase1_QA',
    'qa_spec_date': '2022-03-28',
    'sensor': '*',
    'dtype': {'kind': 'u', 'itemsize': 1},
    'bits': [
        {'bit_number': 0, 'qa_name': 'combined', 'qa_description': 'TnE'},
        {'bit_number': 1, 'qa_name': 'dilated_cloud', 'qa_description': 'dilated_cloud'},
        {'bit_number': 2, 'qa_name': 'cirrus', 'qa_description': 'cirrus'},
        {'bit_number': 3, 'qa_name': 'cloud', 'qa_description': 'cloud'},
        {'bit_number': 4, 'qa_name': 'cloud_shadow', 'qa_description': 'cloud_shadow'},
        {'bit_number': 5, 'qa_name': 'ice', 'qa_description': 'snow'},
        {'bit_number': 6, 'qa_name': 'clear', 'qa_description': 'clear'},
        {'bit_number': 7, 'qa_name': 'water', 'qa_description': 'water'}
    ]
}))


# Sentinel2 L2A SCL scene classification mask
# https://docs.sentinel-hub.com/api/latest/data/sentinel-2-l2a/
QA_SPECS.append(QA_ValueSpecTable({
    'qa_spec_name': 'SCL',
    'qa_spec_date': '???',
    'sensor': 'S2',
    'dtype': {'kind': 'u', 'itemsize': 1},
    'values': [
        {'value': 0, 'qa_name': 'nodata', 'qa_description': 'No data'},
        {'value': 1, 'qa_name': 'defective', 'qa_description': 'Saturated / Defective'},
        {'value': 2, 'qa_name': 'dark_area', 'qa_description': 'Dark Area Pixels'},
        {'value': 3, 'qa_name': 'cloud_shadow', 'qa_description': 'Cloud Shadows'},
        {'value': 4, 'qa_name': 'vegetation', 'qa_description': 'Vegetation'},
        {'value': 5, 'qa_name': 'bare', 'qa_description': 'Bare Soils'},
        {'value': 6, 'qa_name': 'water', 'qa_description': 'Water'},
        {'value': 7, 'qa_name': 'clouds_lo_prob', 'qa_description': 'Clouds low probability / Unclassified'},
        {'value': 8, 'qa_name': 'clouds_mid_prob', 'qa_description': 'Clouds medium probability'},
        {'value': 9, 'qa_name': 'cloud', 'qa_description': 'Clouds high probability'},
        {'value': 10, 'qa_name': 'cirrus', 'qa_description': 'Cirrus'},
        {'value': 11, 'qa_name': 'ice', 'qa_description': 'Snow / Ice'},
    ]
}))


# Landsat Level 2 QA bands
# https://www.usgs.gov/landsat-missions/landsat-collection-2-quality-assessment-bands
# https://www.usgs.gov/landsat-missions/landsat-collection-1-level-1-quality-assessment-band
QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'qa_pixel',
    'qa_spec_date': '???',
    'sensor': 'L8',
    'dtype': {'kind': 'u', 'itemsize': 2},
    'bits': [
        {'bit_number': 0, 'qa_name': 'fill', 'qa_description': 'Designated fill'},
        {'bit_number': 1, 'qa_name': 'terrain_occlusion', 'qa_description': 'Terrain Occlusion'},
        {'bit_number': 4, 'qa_name': 'cloud', 'qa_description': 'Cloud'},
    ],
    # Represents multi-valued groups
    'bit_ranges': [
        {'bit_range': [2, 3], 'qa_name': 'rad_sat', 'qa_description': 'Radiometric Saturation'},
        {'bit_range': [5, 6], 'qa_name': 'cloud_conf', 'qa_description': 'Cloud Confidence'},
        {'bit_range': [7, 8], 'qa_name': 'cloud_shadow_conf', 'qa_description': 'Cloud Shadow Confidence'},
        {'bit_range': [9, 10], 'qa_name': 'snow_conf', 'qa_description': 'Snow/Ice Confidence'},
        {'bit_range': [11, 12], 'qa_name': 'snow_conf', 'qa_description': 'Snow/Ice Confidence'},
    ]
}))


# ARA data
QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'ARA-4',
    'qa_spec_date': '2024-03-06',
    'sensor': 'L8',
    'dtype': {'kind': 'u', 'itemsize': 2},
    'bits': [
        {
            'bit_number': 0,
            'qa_name': 'TnE',
            'qa_description': 'TnE Evaluate',
            'bit_value': [{'value': 1, 'description': 'use-pixel'}, {'value': 0, 'description': 'ignore-pixel'}]
        },
        {
            'bit_number': 1,
            'qa_name': 'dilated_cloud',
            'qa_description': 'cloud',
            'bit_value': [{'value': 1, 'description': 'yes'}, {'value': 0, 'description': 'no'}]
        },
        {'bit_number': 2, 'qa_name': 'cirrus', 'qa_description': 'cirrus'},
        {'bit_number': 3, 'qa_name': 'cloud', 'qa_description': 'cloud'},
        {'bit_number': 4, 'qa_name': 'cloud_shadow', 'qa_description': 'cloud shadow'},
        {'bit_number': 5, 'qa_name': 'snow', 'qa_description': 'snow'},
        {'bit_number': 6, 'qa_name': 'clear', 'qa_description': 'clear'},
        {'bit_number': 7, 'qa_name': 'water', 'qa_description': 'water'},
        {'bit_number': 8, 'qa_name': 'aligned', 'qa_description': 'image aligned'},
        {'bit_number': 11, 'qa_description': 'reserved for future use'},
        {'bit_number': 12, 'qa_description': 'reserved for future use'},
        {'bit_number': 13, 'qa_description': 'reserved for future use'},
        {'bit_number': 14, 'qa_description': 'reserved for future use'},
        {'bit_number': 15, 'qa_description': 'reserved for future use'}
    ]
}))


QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'ARA-4',
    'qa_spec_date': '2024-03-06',
    'sensor': 'S2',
    'dtype': {'kind': 'u', 'itemsize': 2},
    'bits': [
        {'bit_number': 0, 'qa_name': 'TnE', 'qa_description': 'TnE Evaluate', 'bit_value': [{'value': 1, 'description': 'use-pixel'}, {'value': 0, 'description': 'ignore-pixel'}]},
        {'bit_number': 2, 'qa_name': 'cirrus', 'qa_description': 'cirrus'},
        {'bit_number': 3, 'qa_name': 'cloud', 'qa_description': 'cloud'},
        {'bit_number': 4, 'qa_name': 'cloud_shadow', 'qa_description': 'cloud shadow'},
        {'bit_number': 5, 'qa_name': 'snow', 'qa_description': 'snow'},
        {'bit_number': 6, 'qa_name': 'clear', 'qa_description': 'clear'},
        {'bit_number': 7, 'qa_name': 'water', 'qa_description': 'water'},
        {'bit_number': 8, 'qa_name': 'aligned', 'qa_description': 'image aligned'},
        {'bit_number': 10, 'qa_name': 'imputed', 'qa_description': 'cloud imputed'},
        {'bit_number': 11, 'qa_description': 'reserved for future use'},
        {'bit_number': 12, 'qa_description': 'reserved for future use'},
        {'bit_number': 13, 'qa_description': 'reserved for future use'},
        {'bit_number': 14, 'qa_description': 'reserved for future use'},
        {'bit_number': 15, 'qa_description': 'reserved for future use'}
    ]
}))


QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'ARA-4',
    'qa_spec_date': '2024-03-06',
    'sensor': 'PD',
    'dtype': {'kind': 'u', 'itemsize': 2},
    'bits': [
        {'bit_number': 0, 'qa_name': 'TnE', 'qa_description': 'TnE Evaluate', 'bit_value': [{'value': 1, 'description': 'use-pixel'}, {'value': 0, 'description': 'ignore-pixel'}]},
        {'bit_number': 1, 'qa_name': 'dilated_cloud', 'qa_description': 'cloud', 'bit_value': [{'value': 1, 'description': 'yes'}, {'value': 0, 'description': 'no'}]},
        {'bit_number': 3, 'qa_name': 'cloud', 'qa_description': 'cloud'},
        {'bit_number': 6, 'qa_name': 'clear', 'qa_description': 'clear'},
        {'bit_number': 8, 'qa_name': 'aligned', 'qa_description': 'image aligned'},
        {'bit_number': 11, 'qa_description': 'reserved for future use'},
        {'bit_number': 12, 'qa_description': 'reserved for future use'},
        {'bit_number': 13, 'qa_description': 'reserved for future use'},
        {'bit_number': 14, 'qa_description': 'reserved for future use'},
        {'bit_number': 15, 'qa_description': 'reserved for future use'}
    ]
}))


QA_SPECS.append(QA_BitSpecTable({
    'qa_spec_name': 'ARA-4',
    'qa_spec_date': '2024-03-06',
    'sensor': 'WV',
    'dtype': {'kind': 'u', 'itemsize': 2},
    'bits': [
        {'bit_number': 0, 'qa_name': 'TnE', 'qa_description': 'TnE Evaluate', 'bit_value': [{'value': 1, 'description': 'use-pixel'}, {'value': 0, 'description': 'ignore-pixel'}]},
        {'bit_number': 1, 'qa_name': 'dilated_cloud', 'qa_description': 'cloud', 'bit_value': [{'value': 1, 'description': 'yes'}, {'value': 0, 'description': 'no'}]},
        {'bit_number': 3, 'qa_name': 'cloud', 'qa_description': 'cloud'},
        {'bit_number': 3, 'qa_name': 'terrain_shadow', 'qa_description': 'Terrain Shadow'},
        {'bit_number': 5, 'qa_name': 'snow', 'qa_description': 'snow'},
        {'bit_number': 6, 'qa_name': 'clear', 'qa_description': 'clear'},
        {'bit_number': 7, 'qa_name': 'water', 'qa_description': 'water'},
        {'bit_number': 8, 'qa_name': 'aligned', 'qa_description': 'image aligned'},
        {'bit_number': 9, 'qa_name': 'orthorectified', 'qa_description': 'DSM Orthorectified'},
        {'bit_number': 11, 'qa_description': 'reserved for future use'},
        {'bit_number': 12, 'qa_description': 'reserved for future use'},
        {'bit_number': 13, 'qa_description': 'reserved for future use'},
        {'bit_number': 14, 'qa_description': 'reserved for future use'},
        {'bit_number': 15, 'qa_description': 'reserved for future use'}
    ]
}))


def demo():
    """
    Small script to viz qa bands.
    """
    import sys
    fpath = sys.argv[1]
    from geowatch.tasks.fusion.datamodules.qa_bands import QA_SPECS
    # table = QA_SPECS.find_table('ACC-1', 'WV')
    table = QA_SPECS.find_table('ARA-4', 'L8')
    import kwimage
    quality_im = kwimage.imread(fpath, nodata_method='ma')
    drawings = table.draw_labels(quality_im)
    qa_canvas = drawings['qa_canvas']
    legend = drawings['legend']
    canvas = kwimage.stack_images([qa_canvas, legend], axis=1)
    import kwplot
    kwplot.autompl()
    kwplot.imshow(canvas)
    kwplot.show_if_requested()


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/geowatch/tasks/fusion/datamodules/qa_bands.py /home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc/Aligned-Drop7-DEBUG/US_R007/WV/affine_warp/crop_20150401T160000Z_N34.190052W083.941277_N34.327136W083.776956_WV_0/crop_20150401T160000Z_N34.190052W083.941277_N34.327136W083.776956_WV_0_quality.tif

        python ~/code/watch/geowatch/tasks/fusion/datamodules/qa_bands.py /data/joncrall/dvc-repos/smart_phase3_data/Aligned-Drop8-ARA/KR_R001/L8/affine_warp/crop_20140225T010000Z_N37.643680E128.649453_N37.683356E128.734073_L8_0/crop_20140225T010000Z_N37.643680E128.649453_N37.683356E128.734073_L8_0_quality.tif --show
    """
    demo()



"""
Functions that may be moved to kwutil
"""
import itertools as it


def distributed_subitems(items, max_num):
    """
    Return a subset of items maximally distributed over the input index space.
    I.e. the chosen indexes are maximally dialted.

    Args:
        items (List | Dict): an ordered indexable object

    Returns:
        List | Dict: a subset of the input with a length at most ``max_num``.

    TODO:
        - [ ] Find a better name. ChatGPT suggests using "spread", which I
              like. Maybe spreadsort, spreadshuffle, spredtraverse?
              spreadtake, takespread?
        - [ ] Figure out where this lives, probably kwutil.
        - [ ] maybe we should force this to be a generator? Or make a generator variant?


    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> items = list(range(100))
        >>> max_num = 5
        >>> sub_items = distributed_subitems(items, max_num)
        >>> print(sub_items)
        [0, 25, 50, 75, 99]

    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> items = {chr(i): i for i in range(ord('a'), ord('a') + 26)}
        >>> max_num = 5
        >>> sub_items = distributed_subitems(items, max_num)
        >>> print(sub_items)
        {'a': 97, 'g': 103, 'n': 110, 't': 116, 'z': 122}
    """
    sub_index_gen = farthest_from_previous(0, len(items))
    sub_indices = sorted(it.islice(sub_index_gen, max_num))
    if isinstance(items, dict):
        item_keys = list(items.keys())
        sub_keys = [item_keys[idx] for idx in sub_indices]
        sub_items = {key: items[key] for key in sub_keys}
    else:
        sub_items = [items[idx] for idx in sub_indices]
    return sub_items


def farthest_from_previous(start, stop):
    """
    Given a ordered list of items, incrementally yield indexes such that each
    new index maximizes the distance to all other previously chosen indexes.

    Args:
        start (int): The inclusive starting index (typically 0)
        stop (int): The exclusive maximum index (typically ``len(items)``)

    Yields:
        int: the next chosen index in the series

    TODO:
        - [ ] Find a Better Name,
                spread_indices?
                farthest_index_traversal
                spread_index_traversal

        - [ ] Allow the user to specify which point is first?
              Do we start at the beginning, end, or middle?
              Currently we default to the end.

    Notes:
        * This is an instance of farthest-point traversal in 1D.

    References:
        .. [CSSE_167943] https://cs.stackexchange.com/questions/167943/is-this-knapsack-variant-named-studied-online-algorithm-for-farthest-from-pr

    CommandLine:
        xdoctest -m geowatch.utils.util_kwutil farthest_from_previous

    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> total = 10
        >>> start, stop = 0, 10
        >>> gen = farthest_from_previous(start, stop)
        >>> result = list(gen)
        >>> assert set(result) == set(range(start, stop))
        >>> print(result)
        [9, 0, 5, 2, 7, 1, 6, 3, 8, 4]
    """
    if start < stop:
        yield stop - 1
        yield from _farthest_from_previous_helper(start, stop - 1)


def _farthest_from_previous_helper(start, stop):
    if start < stop:
        low_mid: int = (start + stop) // 2
        high_mid: int = (start + stop + 1) // 2

        left_gen = _farthest_from_previous_helper(start, low_mid)
        right_gen = _farthest_from_previous_helper(high_mid, stop)

        pairgen = it.zip_longest(left_gen, right_gen)
        flatgen = it.chain.from_iterable(pairgen)
        filtgen = filter(lambda x: x is not None, flatgen)
        yield from filtgen
        if low_mid < high_mid:
            yield low_mid



"""
Defines :class:`TimeKernel`.

Note:
    Kernel may not be the best name for this, which is a highly overloaded
    term. The name was chosen based on the use of a statistical kernels or
    convolutional kernels used in image processing
    [WikiKernelImageProcessing]_.  It has nothing to do with kernel methods in
    optimization or null space in linear algebra.

    It is really an ideal time distribution that will be used as a template.

References:
    .. [WikiKernelImageProcessing] https://en.wikipedia.org/wiki/Kernel_(image_processing)
"""
import numpy as np


class TimeKernel(np.ndarray):
    """
    Represents an ideal relative time sampling pattern.

    This is just an ndarray with offsets specified in seconds.

    Notes:
        https://numpy.org/doc/stable/user/basics.subclassing.html#extra-gotchas-custom-del-methods-and-ndarray-base

    CommandLine:
        xdoctest -m geowatch.tasks.fusion.datamodules.temporal_sampling.time_kernel TimeKernel --show

    Example:
        >>> from geowatch.tasks.fusion.datamodules.temporal_sampling.time_kernel import TimeKernel
        >>> from geowatch.tasks.fusion.datamodules.temporal_sampling.time_kernel import _random_discrete_relative_times
        >>> self = TimeKernel.coerce('-1y,-3m,0,3m,1y')
        >>> print(f'self={self!r}')
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autosns()
        >>> # Pretend we have discrete observations with the following relative
        >>> # time differences
        >>> import kwutil
        >>> time_range = kwutil.timedelta.coerce('4y')
        >>> relative_unixtimes = _random_discrete_relative_times(time_range)
        >>> self.plot(relative_unixtimes)
        >>> kwplot.show_if_requested()
    """
    def __new__(cls, *args, **kwargs):
        print('In __new__ with class %s' % cls)
        return super().__new__(cls, *args, **kwargs)

    @classmethod
    def coerce(cls, data):
        from geowatch.tasks.fusion.datamodules.temporal_sampling.utils import coerce_time_kernel
        kernel = np.array(coerce_time_kernel(data))
        self = kernel.view(cls)
        return self

    @classmethod
    def coerce_multiple(cls, data):
        """
        Example:
            >>> from geowatch.tasks.fusion.datamodules.temporal_sampling.time_kernel import TimeKernel
            >>> import ubelt as ub
            >>> pattern = ('-3y', '-2.5y', '-2y', '-1.5y', '-1y', 0, '1y', '1.5y', '2y', '2.5y', '3y')
            >>> multi_kernel = TimeKernel.coerce_multiple(pattern)
            >>> print('multi_kernel = {}'.format(ub.urepr(multi_kernel, nl=2)))
        """
        from geowatch.tasks.fusion.datamodules.temporal_sampling.utils import coerce_multi_time_kernel
        kernels = coerce_multi_time_kernel(data)
        kernels = [np.array(k).view(cls) for k in kernels]
        return kernels

    def make_soft_mask(self, relative_unixtimes):
        from geowatch.tasks.fusion.datamodules.temporal_sampling.affinity import make_soft_mask
        kernel_masks, kernel_attrs = make_soft_mask(self, relative_unixtimes)
        return kernel_masks, kernel_attrs

    def plot(self, relative_unixtimes):
        time_kernel = self
        kernel_masks, kernel_attrs = self.make_soft_mask(relative_unixtimes)

        min_t = min(kattr['left'] for kattr in kernel_attrs)
        max_t = max(kattr['right'] for kattr in kernel_attrs)

        min_t = min(min_t, relative_unixtimes[0])
        max_t = max(max_t, relative_unixtimes[-1])

        import kwplot
        from geowatch.utils import util_kwplot
        plt = kwplot.autoplt()
        import kwimage
        kwplot.close_figures()
        kwplot.figure(fnum=1, doclf=1)
        kernel_color = kwimage.Color.coerce('kitware_green').as01()
        obs_color = kwimage.Color.coerce('kitware_darkblue').as01()

        kwplot.figure(fnum=1, pnum=(1, 1, 1))

        kwplot.phantom_legend({
            'Ideal Sample': kernel_color,
            'Discrete Observation': obs_color,
        })

        for kattr in kernel_attrs:
            rv = kattr['rv']
            xs = np.linspace(min_t, max_t, 1000)
            ys = rv.pdf(xs)
            kattr['_our_norm'] = ys.sum()
            ys_norm = ys / ys.sum()
            plt.plot(xs, ys_norm)

        ax = plt.gca()
        # ax.set_ylim(0, 1)
        ax.set_xlabel('relative time (days)')
        ax.set_ylabel('sample probability')
        # ax.set_title('ideal sample location')
        ax.set_yticks([])

        lw = 2

        obs_line_segments = []
        for x in relative_unixtimes:
            y = 0
            for kattr in kernel_attrs:
                rv = kattr['rv']
                y = max(y, rv.pdf(x) / kattr['_our_norm'])
            obs_line_segments.append([x, y])
        for x, y in obs_line_segments:
            plt.plot([x, x], [0, y], '-', color=obs_color, linewidth=lw)

        kern_line_segments = []
        for x in time_kernel:
            y = 0
            for kattr in kernel_attrs:
                rv = kattr['rv']
                y = max(y, rv.pdf(x) / kattr['_our_norm'])
            kern_line_segments.append([x, y])
        for x, y in kern_line_segments:
            plt.plot([x, x], [0, y], '--', color=kernel_color)

        # plt.plot(time_kernel, [0] * len(time_kernel), '-o', color=kernel_color, label='ideal frame location')

        kwplot.phantom_legend(label_to_attrs={
            'Ideal Sample': {'color': kernel_color, 'linestyle': '--'},
            'Discrete Observation': {'color': obs_color, 'linewidth': lw},
        })

        util_kwplot._format_xaxis_as_timedelta(ax)

        # plt.subplots_adjust(top=0.99, bottom=0.1, hspace=.3, left=0.1)
        # fig = plt.gcf()
        # fig.set_size_inches(np.array([4, 3]) * 1.5)
        # fig.tight_layout()
        # finalizer = util_kwplot.FigureFinalizer()
        # finalizer(fig, 'time_sampling_example.png')


def _random_discrete_relative_times(time_range):
    import kwarray
    rng = kwarray.ensure_rng()
    relative_unixtimes = rng.rand(10) * time_range.total_seconds()
    idx = ((relative_unixtimes - (time_range.total_seconds() / 2)) ** 2).argmin()
    mid = relative_unixtimes[idx]
    relative_unixtimes = relative_unixtimes - mid
    return relative_unixtimes
